@@ -23,6 +23,9 @@ typedef struct {
     ptable_ptr ref_bless_av;            /* ptr table for tracking which objects need to be bless - key: ofs, value: mortal AV (of refs)  */
     AV* weakref_av;
 
+    AV* alias_cache; /* used to cache integers of different sizes. */
+    UV alias_varint_under;
+
     UV bytes_consumed;
     UV recursion_depth;                 /* Recursion depth of current decoder */
     U8 proto_version_and_flags;
@@ -75,7 +78,7 @@ void srl_decoder_destructor_hook(pTHX_ void *p);
 #define SRL_ERRORf3(fmt,var1,var2,var3)         croak(SRL_BASE_ERROR_FORMAT fmt, SRL_BASE_ERROR_ARGS, (var1),(var2),(var3))
 #define SRL_ERRORf4(fmt,var1,var2,var3,var4)    croak(SRL_BASE_ERROR_FORMAT fmt, SRL_BASE_ERROR_ARGS, (var1),(var2),(var3),(var4))
 #define SRL_ERROR_UNIMPLEMENTED(dec,tag,str) \
-    SRL_ERRORf3("Tag %u %s is unimplemented at ofs: %d", (tag), (str), BUF_POS_OFS(dec))
+    SRL_ERRORf3("Tag %u %s is unimplemented at ofs: %lu", (tag), (str), (unsigned long)BUF_POS_OFS(dec))
 #define SRL_ERROR_UNTERMINATED(dec,tag,str) \
     SRL_ERRORf4("Tag SRL_HDR_%s %s was not terminated properly at ofs %lu with %lu to go", \
             tag_name[(tag) & 127], (str), (dec)->pos - (dec)->buf_start, (dec)->buf_end - (dec)->pos)
@@ -89,31 +92,44 @@ void srl_decoder_destructor_hook(pTHX_ void *p);
 
 /* If set, the decoder struct needs to be cleared instead of freed at
  * the end of a deserialization operation */
-#define SRL_F_REUSE_DECODER 1UL
+#define SRL_F_REUSE_DECODER                     0x00000001UL
 /* If set, then the decoder is already in use and srl_decode_into will
  * clone its own new decoder. */
-#define SRL_F_DECODER_DIRTY 2UL
+#define SRL_F_DECODER_DIRTY                     0x00000002UL
 /* Non-persistent flag! */
-#define SRL_F_DECODER_NEEDS_FINALIZE 4UL
+#define SRL_F_DECODER_NEEDS_FINALIZE            0x00000004UL
 /* Non-persistent flag! */
-#define SRL_F_DECODER_DECOMPRESS_SNAPPY 8UL
-/* Persistent flag: Make the decoder REFUSE compressed documents */
-#define SRL_F_DECODER_REFUSE_SNAPPY 16UL
+#define SRL_F_DECODER_DECOMPRESS_SNAPPY         0x00000008UL
+/* Non-persistent flag! */
+#define SRL_F_DECODER_DECOMPRESS_ZLIB           0x00000010UL
+/* Persistent flag: Make the decoder REFUSE Snappy-compressed documents */
+#define SRL_F_DECODER_REFUSE_SNAPPY             0x00000020UL
+/* Persistent flag: Make the decoder REFUSE zlib-compressed documents */
+#define SRL_F_DECODER_REFUSE_ZLIB               0x00000040UL
 /* Persistent flag: Make the decoder REFUSE objects */
-#define SRL_F_DECODER_REFUSE_OBJECTS 32UL
+#define SRL_F_DECODER_REFUSE_OBJECTS            0x00000080UL
 /* Persistent flag: Make the decoder validate UTT8 strings */
-#define SRL_F_DECODER_VALIDATE_UTF8 64UL
+#define SRL_F_DECODER_VALIDATE_UTF8             0x00000100UL
 /* Persistent flag: Make the encoder forget to bless */
-#define SRL_F_DECODER_NO_BLESS_OBJECTS 128UL
+#define SRL_F_DECODER_NO_BLESS_OBJECTS          0x00000200UL
 /* Persistent flag: Destructive incremental parsing */
-#define SRL_F_DECODER_DESTRUCTIVE_INCREMENTAL 256UL
+#define SRL_F_DECODER_DESTRUCTIVE_INCREMENTAL   0x00000400UL
 /* Non-persistent flag: The current packet is using protocol version 1 */
-#define SRL_F_DECODER_PROTOCOL_V1 512UL
+#define SRL_F_DECODER_PROTOCOL_V1               0x00000800UL
+/* Persistent flag: alias small integer values in Hashes and Arrays */
+#define SRL_F_DECODER_ALIAS_SMALLINT            0x00001000UL
+/* Persistent flag: use PL_sv_undef for undef values in Hashes and Arrays */
+#define SRL_F_DECODER_ALIAS_VARINT              0x00002000UL
+/* Persistent flag: use PL_sv_undef as many places as possible */
+#define SRL_F_DECODER_USE_UNDEF                 0x00004000UL
+
+
+#define SRL_F_DECODER_ALIAS_CHECK_FLAGS   ( SRL_F_DECODER_ALIAS_SMALLINT | SRL_F_DECODER_ALIAS_VARINT | SRL_F_DECODER_USE_UNDEF )
 
 #define SRL_DEC_HAVE_OPTION(dec, flag_num) ((dec)->flags & flag_num)
 #define SRL_DEC_SET_OPTION(dec, flag_num) ((dec)->flags |= flag_num)
 #define SRL_DEC_UNSET_OPTION(dec, flag_num) ((dec)->flags &= ~flag_num)
-#define SRL_DEC_VOLATILE_FLAGS (SRL_F_DECODER_NEEDS_FINALIZE|SRL_F_DECODER_DECOMPRESS_SNAPPY|SRL_F_DECODER_PROTOCOL_V1|SRL_F_DECODER_DIRTY)
+#define SRL_DEC_VOLATILE_FLAGS (SRL_F_DECODER_NEEDS_FINALIZE|SRL_F_DECODER_DECOMPRESS_SNAPPY|SRL_F_DECODER_PROTOCOL_V1|SRL_F_DECODER_DIRTY|SRL_F_DECODER_DECOMPRESS_ZLIB)
 #define SRL_DEC_RESET_VOLATILE_FLAGS(dec) ((dec)->flags &= ~SRL_DEC_VOLATILE_FLAGS)
 
 /* 
@@ -179,7 +195,7 @@ static const char * const tag_name[] = {
 	"RESERVED_2",        /* "6"   54 0x36 0b00110110 */
 	"RESERVED_3",        /* "7"   55 0x37 0b00110111 */
 	"RESERVED_4",        /* "8"   56 0x38 0b00111000 */
-	"RESERVED_5",        /* "9"   57 0x39 0b00111001 */
+	"SV_UNDEF",          /* "9"   57 0x39 0b00111001 */
 	"FALSE",             /* ":"   58 0x3a 0b00111010 */
 	"TRUE",              /* ";"   59 0x3b 0b00111011 */
 	"MANY",              /* "<"   60 0x3c 0b00111100 */
